import collections
import os
import sys
import tempfile
import unittest
import subprocess
import utils
import re

"""class CompileArgsTestCase(unittest.TestCase):
    def test(self):
        args = CompileArgs(
            source = 'source.cpp',
            output = 'output.out',
            link = False,
            strict = True,
        )
        
        self.assertEqual(list(args), ['g++', 'source.cpp', '-o', 'output.out',
                                      '-c', '-Wall'])
"""
class BaseCompilerTestCase(unittest.TestCase):
    def test_join(self):
        base = BaseCompiler(directory='/testing/')
        self.assertEqual(base.join('is.fun'), '/testing/is.fun')
    
    def test_string_includes(self):
        base = BaseCompiler(includes=['testing/is/the/best.cpp',
                                      'everyone/should/test.cpp'])
        
        string_includes = base.get_context()['includes']
        self.assertIn('best.h', string_includes)
        self.assertIn('test.h', string_includes)
        
    def test_brace_validation(self):
        base = BaseCompiler()

        self.assertEqual(base.check_braces("{ char blah = '3\"; }") == None, False)
        self.assertEqual(base.check_braces("{ char blah = '3'; }") == None, True)
        self.assertEqual(base.check_braces("{ const char* blah = \"abcdefg'; }") == None, False)
        self.assertEqual(base.check_braces("{ ") == None, False)
        self.assertEqual(base.check_braces("{ \" [] () \" }") == None, True)

class CompileArgs(object):
    def __init__(self, source, output=None, includes=None, link=True, strict=False):
        self.source = source
        self.output = output
        if includes is None:
            includes = []
        self.includes = includes
        self.link = link
        self.strict = strict
    
    def __iter__(self):
        """Pretend that this is a list."""
        from core import function
        
        # Compile the source file.
        args = ['g++']
        
        if self.strict:
            args.extend(['-Werror', '-Wall',])
        
        args.append(self.source)

        # Include the source files for each of the included functions.
        if self.link:
            for include in self.includes:
                if isinstance(include, str):
                    args.append('%s.cpp' % include.rsplit('.', 1)[0])
                else:
                    args.append(include.get_source_path())
        
        # Include the directories containing each of the included functions in
        # the search path.
        for include in self.includes:
            if isinstance(include, str):
                args.extend(['-I', os.path.dirname(include)])
            else:
                args.extend(['-I', include.directory])
        
        # Supply an output argument if we were given a path.
        if self.output:
            args.extend(['-o', self.output])

        # We're not supposed to link the files (e.g. because they're functions),
        # add the required parameter.
        if not self.link:
            args.append('-c')
        
        
        return iter(args)
            

class BaseCompiler(object):
    def __init__(self, includes=None, directory=None, strict=False, gory=False):
        if directory is None:
            directory = tempfile.mkdtemp()
        self.directory = directory
        
        self.strict = strict
        self.gory = gory
        if includes is None:
            includes = []
        self.includes = includes
            
    def join(self, *args):
        """Appends the given path to the predefined directory."""
        return os.path.join(self.directory, *args)
        
    def get_source_name(self):
        """Returns the name of the .cpp file."""
        return '%s.cpp' % self.name
    
    def get_source_path(self):
        """Returns the full path to the .cpp file."""
        return self.join(self.get_source_name())
    
    def get_compiled_name(self):
        """Returns the name of the compiled/executable file."""
        if sys.platform.startswith("win"):
            return '%s.exe' % self.name
        return '%s.out' % self.name
    
    def get_compiled_path(self):
        """Returns the full path to the compiled/executable file."""
        return self.join(self.get_compiled_name())

    def render_to_string(self, template, context):
        """Renders the template at the given path using the supplied
        context dictionary and returns it as a string.
        
        """
        
        # Setup a defaultdict for the context, so that all nonexistent keys
        # default to empty strings.
        defaultcontext = collections.defaultdict(str)
        defaultcontext.update(context)
        
        # Read and render the template.
        template_path = os.path.join(utils.ROOT, 'templates', template)
        with open(template_path, "r") as fd:
            return fd.read() % defaultcontext
    
    def render_to_file(self, template, context, output):
        """Renders the template at the given path using the supplied
        context dictionary and writes it to the specified path.
        
        """
       
        # Write the rendered template to the given output path.
        with open(output, 'w') as fd:
            fd.write(self.render_to_string(template, context))
        
        # Return the path of the file.
        return fd.name        
    
    def get_context(self):
        """Returns a dictionary containing all of the values that can be
        inserted into templates.
        
        """
        from core import function
        
        include_statements = []
        
        for include in self.includes:
            if isinstance(include, function.FunctionCompiler):
                include_statements.append('#include "%s.h"' % include.name)
            else:
                include_statements.append('#include "%s.h"' %
                    os.path.basename(include).rsplit('.', 1)[0])
        
                              
        return {'includes': '\n'.join(include_statements)}
    
    def get_compile_args(self):
        """Returns a list of arguments that should be passed to
        subprocess.check_output in order to compile the file.
        
        """
        
        return CompileArgs(
            source = self.get_source_path(),
            output = self.join(self.get_compiled_name()),
            includes = self.includes,
            strict = self.strict,
        )
    
    def write_source_files(self):
        """Writes the source files to disk."""
        raise NotImplementedError

    def check_braces(self, originaltext):
        """ Makes sure that braces and quotes are balanced and returns a string describing the problem if they aren't"""
        
        
        
        braceTypes = { '[' : ']',
                       '{' : '}',
                       '(' : ')'
                       }
       
        testtext = originaltext.replace(r'\\','')
        testtext = testtext.replace(r'\"','')
        testtext = testtext.replace(r"\'","")

        pattern = re.compile(r"""("(.|\n)*?")|('(.|\n)*?')|(/\*(.|\n)*?\*/)|(//.*?\n)""")
        testtext = pattern.sub('', testtext)
        

        if testtext.count("'") % 2 != 0:
            return "Mismatched single quotes in the body."
        
        if testtext.count('"') % 2 != 0:
            return "Mismatched double quotes in the body."

        
        for brace in braceTypes.keys():
            if(testtext.count(brace) != testtext.count(braceTypes[brace])):
                return "Mismatched " + brace + braceTypes[brace] + " in the body."
            
        return None
    
    def compile(self):
        """Renders the class's templates, and calls g++ to compile them."""
        
        self.write_source_files()
        kwargs = {'stderr': subprocess.STDOUT}
        

        # Add the location of g++ to the system path if we are running windows.
        if subprocess.mswindows:
            locations = os.environ['PATH'].split(';')
            if r"C:\dev-cpp\bin" not in locations:
                locations.append(r"C:\dev-cpp\bin")
                os.environ['PATH'] = ';'.join(locations)
            
            kwargs['creationflags'] = 0x08000000

        compile_args = self.get_compile_args()
        
        # Compile the source file using the arguments from above.
        output = subprocess.check_output(compile_args, **kwargs)
