import collections
import os
import sys
import re
import tempfile
import unittest
import subprocess
import utils
from core import base

class ExpressionCompilerTestCase(unittest.TestCase):
    def evaluate(self, expression):
        exp = ExpressionCompiler(expression)
        return exp.evaluate()
    
    def test_addition(self):
        self.assertEqual(self.evaluate("2 + 2"), '4')
    
    def test_subtraction(self):
        self.assertEqual(self.evaluate("100 - 314"), '-214')

    def test_multiplication(self):
        self.assertEqual(self.evaluate("2 * 3"), '6')

    def test_division(self):
        self.assertEqual(self.evaluate("7 / 3"), '2')

    def test_squareroot(self):
        self.assertEqual(self.evaluate("sqrt(4)"), '2')
    
    def test_boolean_or(self):
        self.assertEqual(self.evaluate("true || false"), 'true')
    
    def test_boolean_and(self):
        self.assertEqual(self.evaluate("true && false"), 'false')

    def test_equality(self):
        self.assertEqual(self.evaluate("5 == 5"), 'true')

    def test_inequality(self):
        self.assertEqual(self.evaluate("5 != 11"), 'true')
        
    def test_greater_than(self):
        self.assertEqual(self.evaluate("5 > 3"), 'true')
        
    def test_less_than(self):
        self.assertEqual(self.evaluate("5 < 3986"), 'true')
    
    def test_name(self):
        exp = ExpressionCompiler('1+1')
        self.assertEqual(exp.name, 'expression')

    def test_source_name(self):
        exp = ExpressionCompiler('1+1')
        self.assertEqual(exp.get_source_name(), 'expression.cpp')

    def test_source_path(self):
        exp = ExpressionCompiler('1+1', directory='/testing/rocks/')
        self.assertEqual(exp.get_source_path(), '/testing/rocks/expression.cpp')

    def test_compiled_name(self):
        exp = ExpressionCompiler('1+1')
        if sys.platform.startswith("win"):
            self.assertEqual(exp.get_compiled_name(), 'expression.exe')
        else:
            self.assertEqual(exp.get_compiled_name(), 'expression.out')

    def test_compiled_path(self):
        exp = ExpressionCompiler('1+1', directory='/testing/rocks/')
        if sys.platform.startswith("win"):
            self.assertEqual(exp.get_compiled_path(),
                             '/testing/rocks/expression.exe')
        else:
            self.assertEqual(exp.get_compiled_path(),
                             '/testing/rocks/expression.out')

    def test_strict(self):
        exp = ExpressionCompiler('(unsigned char)1 < 0', strict = True)
        self.assertEqual(exp.strict, True)

    def test_gory(self):
        exp = ExpressionCompiler('hi+5', gory = True)
        self.assertEqual(exp.gory, True)
    

class ExpressionCompiler(base.BaseCompiler):
    name = 'expression'
    
    def __init__(self, expression, *args, **kwargs):
        self.expression = expression
        
        super(ExpressionCompiler, self).__init__(*args, **kwargs)

    def get_context(self):
        """Get the variables that should be passed to the templates."""
        context = super(ExpressionCompiler, self).get_context()
        context.update({
            'expression': self.expression,
        })
        return context

    def write_source_files(self):
        """Write the source files to disk."""
        context = self.get_context()
        name = self.name
        
        self.render_to_file('expression.cpp', context,
                            self.join('%s.cpp' % name))

    def evaluate(self):
        """Evaluates the supplied expression."""
        
        try:
            self.compile()
        except OSError as err:
            return "Error: %s" % err
        except subprocess.CalledProcessError as err:
            if self.gory:
                result = err.output.decode('utf-8').splitlines()
                result = '\n'.join([re.sub(r'^[^:]*(\\|/)', '', line).strip() for line in result])
                return result

                error = err.output
                error_list = str(error).splitlines()
                joined_output = '\r\n'.join(error_list)
                return joined_output
            return "Bad expression"
		
        kwargs = {}

        # Add the location of g++ to the system path if we are running windows.
        if subprocess.mswindows:
            kwargs['creationflags'] = 0x08000000

        # Read the result of the evaluated expression from STDOUT.
        try:
            result = subprocess.check_output([self.join(self.get_compiled_name())], **kwargs)
        except subprocess.CalledProcessError as err:
            result = err.output.decode('utf-8').splitlines()
            result = '\n'.join([line.split(':', 1)[1].strip() for line in result])
            return "Expression compiled, but crashed during execution.\n\n" + result
        
        # Decode the raw byte string into actual text. Note that we need to use
        # latin-1 here for the y umlat character, from (unsigned char) -1.
        return result.decode("latin-1").strip()
              
