import collections
import os
import sys
import re
import tempfile
import unittest
import subprocess
import utils
import tempfile
from core import base, expression

class FunctionCompilerTestCase(unittest.TestCase):
    def evaluate(self, *args, **kwargs):
        exp = expression.ExpressionCompiler(*args, **kwargs)
        return exp.evaluate()

    def test_function_include_function_roundtrip(self):
        directory = tempfile.mkdtemp()
        
        func1 = FunctionCompiler(data=dict(
            name = 'add',
            signature = 'add: int int -> int',
            purpose = 'Expects two integers and produces their \nsum.',
            header = 'int add (int val1, int val2)',
            tests = 'add(3,5)==8;add(9,14)==23;',
            constants = 'const int numberOfBearsInUniverse = 42;',
            body = '{return val1 + val2;}',
        ), directory=directory)
        func2 = FunctionCompiler(data=dict(
            name = 'triple',
            signature = 'triple: int -> int',
            purpose = 'Doubles\nan\ninteger!',
            header = 'int triple (int val1)',
            tests = 'triple(1)==3;triple(9)==27;',
            constants = 'const int numberOfTurtlesInUniverse = 31;',
            body = '{return add(val1, val1) + val1;}',
        ), includes=[func1], directory=directory)
        func1.write_source_files()
        func2.write_source_files()
        func1.compile()
        func2.compile()
        
        results = func2.run_tests()
        for result in list(zip(*results))[-1]:
            self.assertEqual(result, 'true')
        
        compilers = load_functions(func2.directory)
        self.assertEqual(len(compilers), 2)
        
        func2copy=[c for c in compilers if c.name == "triple"][0]
        
        self.assertEqual(
            func2copy.render_to_string('function.cpp', func2copy.get_context()),
            func2.render_to_string('function.cpp', func2.get_context()),
        )

        self.assertEqual(
            func2copy.render_to_string('function.h', func2copy.get_context()),
            func2.render_to_string('function.h', func2.get_context()),
        )

    def test_function_roundtrip(self):
        func1 = FunctionCompiler(data=dict(
            name = 'add',
            signature = 'add: int int -> int',
            purpose = 'Expects two integers and produces their \nsum.',
            header = 'int add (int val1, int val2)',
            tests = 'add(3,5)==8;add(9,14)==23;',
            constants = 'const int numberOfBearsInUniverse = 42;',
            body = '{return val1 + val2;}',
        ))
        func1.write_source_files()
        compilers = load_functions(func1.directory)
        self.assertEqual(len(compilers), 1)
        
        func2=compilers[0]
        
        self.assertEqual(
            func1.render_to_string('function.cpp', func1.get_context()),
            func2.render_to_string('function.cpp', func2.get_context()),
        )

        self.assertEqual(
            func1.render_to_string('function.h', func1.get_context()),
            func2.render_to_string('function.h', func2.get_context()),
        )

    def test_function_partial_input_roundtrip(self):
        func1 = FunctionCompiler(data=dict(
            name = 'dude',
            signature = '',
            purpose = '',
            header = '',
            tests = '',
            constants = '',
            body = '{}',
        ))
        func1.write_source_files()
        compilers = load_functions(func1.directory)
        self.assertEqual(len(compilers), 1)
        
        func2=compilers[0]
        
        self.assertEqual(
            func1.render_to_string('function.cpp', func1.get_context()),
            func2.render_to_string('function.cpp', func2.get_context()),
        )

        self.assertEqual(
            func1.render_to_string('function.h', func1.get_context()),
            func2.render_to_string('function.h', func2.get_context()),
        )
    
    def test_function_constants(self):
        func = FunctionCompiler(data=dict(
            name = 'add',
            signature = 'add: int int -> int',
            purpose = 'Expects two integers and produces their sum.',
            header = 'int add (int val1, int val2)',
            constants = 'const int numberOfBearsInUniverse = 42;',
            tests = 'add(3,5)==8;add(9,14)==23;',
            body = '{return val1 + val2;}',
        ))
        func.compile()
        
        self.assertEqual(self.evaluate('numberOfBearsInUniverse == 42',
                                       includes=[func]), 'true')
        
    def test_function_add(self):
        func = FunctionCompiler(data=dict(
            name = 'add',
            signature = 'add: int int -> int',
            purpose = 'Expects two integers and produces their sum.',
            header = 'int add (int val1, int val2)',
            constants = 'const int numberOfBearsInUniverse = 42;',
            tests = 'add(3,5)==8;add(9,14)==23;',
            body = '{return val1 + val2;}',
        ))
        func.compile()
        
        self.assertEqual(self.evaluate('add(7,18)', includes=[func]), '25')
        
        # Get the results from running the functions built-in tests, and assert
        # that they all passed.
        results = func.run_tests()
        for result in list(zip(*results))[-1]:
            self.assertEqual(result, 'true')
    
    def test_function_add_string(self):
        func = FunctionCompiler(data=dict(
            name = 'add',
            signature = 'add: int int -> int',
            purpose = 'Expects two integers and produces their sum.',
            header = 'int add (int val1, int val2)',
            constants = 'const int numberOfBearsInUniverse = 42;',
            tests = 'add(3,5)==8;add(9,14)==23;',
            body = '{return val1 + val2;}',
        ))
        func.compile()
        
        self.assertEqual(self.evaluate('add(7,18)',
                         includes=[func.get_source_path()]), '25')
        
        # Get the results from running the functions built-in tests, and assert
        # that they all passed.
        results = func.run_tests()
        for result in list(zip(*results))[-1]:
            self.assertEqual(result, 'true')
    
    def test_function_strict(self):
        func = FunctionCompiler(data=dict(
            name = 'add',
            signature = 'add: int int -> int',
            purpose = 'Expects two integers and produces their sum.',
            header = 'int add (int val1, int val2)',
            tests = 'add(3,5)==8;add(9,14)==23;',
            body = '{(((unsigned char)1) < 0); return val1 + val2;}',
        ), strict = True)
        self.assertRaises(subprocess.CalledProcessError, func.compile)
        
class FunctionCompiler(base.BaseCompiler):
    def __init__(self, *args, **kwargs):
        
        self.data = kwargs.pop('data', {})
        self.files = kwargs.pop('files', [])
        
        super(FunctionCompiler, self).__init__(*args, **kwargs)
    
    def update(self, *args, **kwargs):
        self.data.update(*args, **kwargs)
    
    def __getattr__(self, name):
        """Alias instance attributes to data keys if they exist."""
        return self.data.get(name, '')
    
    def get_compiled_name(self):
        """Returns the name of the compiled/executable file."""
        return '%s.o' % self.name

    def get_context(self):
        """Get the variables that should be passed to the templates."""
        context = super(FunctionCompiler, self).get_context()
        context.update(self.data)
        context.update(NAME=self.name.upper())
        return context

    def write_source_files(self):
        """Write the source files to disk."""
        context = self.get_context()
        name = self.name
        assert name, "Cannot save a function without a name!"
        
        # Deletes the existing files before writing new ones.
        self.delete()
        self.files = [
            self.render_to_file('function.cpp', context, self.join('%s.cpp' % name)),
            self.render_to_file('function.h', context, self.join('%s.h' % name)),
        ]
        
    def compile(self, *args, **kwargs):
        result = super(FunctionCompiler, self).compile(*args, **kwargs)
        if self.get_compiled_path() not in self.files:
            self.files.append(self.get_compiled_path())
        return result
    
    def get_compile_args(self):
        """Get a list of arguments to be passed to the compiler."""
        args = super(FunctionCompiler, self).get_compile_args()
        args.link = False
        return args
    
    def run_tests(self):
        """Run the user entered tests."""
        results = []
        
        tests = [test.strip() for test in
                 self.tests.replace(';', '\n').splitlines() if test.strip()]
        
        for test in tests:
            includes = list(self.includes) + [self]
            exp = expression.ExpressionCompiler(test, includes=includes)
            passed = exp.evaluate().splitlines()[-1] == 'true' and 'true' or 'false'
            results.append((test, passed))
        
        return results

    def is_editable(self):
        """Returns whether this function can be edited."""
        return True
    
    def is_valid(self):
        """Returns whether this function is already compiled."""
        try:
            compiled_mtime = os.stat(self.get_compiled_path()).st_mtime
            cpp_mtime = os.stat(self.join("%s.cpp" % self.name)).st_mtime
            h_mtime = os.stat(self.join("%s.h" % self.name)).st_mtime
            
            if compiled_mtime < cpp_mtime:
                return False
            if compiled_mtime < h_mtime:
                return False
            return True

        except:
            return False
    
    def delete(self):
        """Deletes any saved source files from the disk."""
        for path in self.files:
            try:
                os.remove(path)
            except OSError:
                pass


class ExternalCompiler(object):
    def __init__(self, name, files):
        self.name = name
        self.files = [file for file in files if
                      file.rsplit('.', 1)[-1] in ('cpp', 'h', 'o')]
    
    def is_editable(self):
        """Whether or not this type of function can be edited."""
        False
    
    def is_valid(self):
        """Returns whether or not the files are already compiled."""
        extensions = dict([(file.rsplit('.', 1)[-1], file) for
                           file in self.files])
        
        if 'o' not in extensions:
            return False
        
        o_mtime = os.stat(extensions['o']).st_mtime
        
        if 'cpp' in extensions:
            cpp_mtime = os.stat(extensions['cpp']).st_mtime
            if cpp_mtime > o_mtime:
                return False
            
        if 'h' in extensions:
            h_mtime = os.stat(extensions['h']).st_mtime
            if h_mtime > o_mtime:
                return False
        
        return True
    
    def delete(self):
        """Remove the files from disk."""
        for file in self.files:
            os.remove(file)

def parse_cpp(path):
    """Parse the values out of a .cpp file."""
    with open(path, "r") as fd:
       readdata = fd.read()

    regex = re.compile(r"""
    /\*USER\ INCLUDES\*/\n*              # The includes statement.
    (?P<includes>(\#include .+\n)*)\n*   # The actual includes.

    /\*\n
    Name:\ (?P<name>.*)\n
    Signature:\ (?P<signature>.*)\n
    Purpose:\ (?P<purpose>(.+\n)*)\n*
    
    Tests:\n
    (?P<tests>(.*\n)*)\n*
    \*/\n*

    /\*HEADER\*/\n
    (?P<header>.*)\n
    /\*BODY\*/\n
    (?P<body>(.*\n)*)
    """, re.I + re.X)
    
    result = regex.search(readdata)
    if result:
        data = result.groupdict().copy()
        data['body'] = data['body'][:-1]
        data['tests'] = data['tests'][:-1]
        data['purpose'] = data['purpose'][:-1]
        
        includes = []
        for line in data['includes'].splitlines():
            include = re.search(r'''^#include "(\w+)\.\w+"''', line)
            if include:
                includes.append(include.group(1))
            
        data['includes'] = includes
        
        return data
    
    raise ValueError('File could not be parsed')

def parse_h(path):
    """Parse the values out of a .h file."""
    with open(path, "r") as fd:
       readdata = fd.read()

    regex = re.compile(r"""
    /\*HEADER\*/\n
    (.+);\n*
    /\*CONSTANTS\*/\n
    (?P<constants>(.+\n)*)\n*
    \#endif
    """, re.I + re.X)
    
    result = regex.search(readdata)
    if result:
        data = result.groupdict().copy()
        data['constants'] = data['constants'][:-1]
        
        return data

    return {}

def load_functions(path):
    """Return a list of Compiler objects for the files in the given path."""
    functions = {}
    for filename in os.listdir(path):
        # Ignore any directories or extensionless files.
        if '.' not in filename:
            continue
        
        basename, extension = filename.rsplit('.', 1)
        if extension not in ('cpp', 'h', 'o'):
            continue
        
        related_files = functions.setdefault(basename, [])
        related_files.append(extension)
    
    compilers = {}
    for name, extensions in functions.items():
        files = [os.path.join(path, "%s.%s" % (name, ext)) for ext in extensions]
        
        if 'cpp' in extensions:
            try:
                data = parse_cpp(os.path.join(path, '%s.cpp' % name))
                
                if 'h' in extensions:
                    try:
                        extra_data = parse_h(os.path.join(path, '%s.h' % name))
                        data.update(extra_data)
                    except ValueError:
                        pass
                                
                # Save the compiler and go on to the next file.
                includes = data.pop('includes', [])
                compiler = FunctionCompiler(data=data, directory=path,
                                            files=files, includes=includes)
            except ValueError:
                continue
                compiler = ExternalCompiler(name, files)
        
        elif 'o' in extensions:
            continue
            compiler = ExternalCompiler(name, files)
        else:
            # The required files do not exist to create a compiler object. So we
            # skip this files and go onto the next. Boom!
            continue
        
        # Add the compiler to the list of all super compiling objects. Batman!
        compilers[compiler.name] = compiler
        
    for compiler in compilers.values():
        compiler.includes = [compilers[name] for name in compiler.includes
                             if name in compilers]
    
    return list(compilers.values())
