from copy import deepcopy

from tkinter import Label, Button, Checkbutton, Entry, IntVar, Frame, Canvas, Scrollbar
from tkinter import NORMAL, DISABLED, END, VERTICAL

from gui.widgets.highlightable import HighlightableList, HighlightableRow
from gui.widgets.smart import SmartButton, SmartCheckbutton
from core.function import FunctionCompiler, ExternalCompiler

from gui.functions.functionrow import Function

from functionGUI import functionWindow 

class FunctionList(HighlightableList):
    def __init__(self, *args, **kwargs):
        self.directory = kwargs.pop('directory', None)
        
        super(FunctionList, self).__init__(*args, **kwargs)
        
        self.new_button = SmartButton(self, name='newbtn', text='New', command=self.new, enabled=True)
        self.edit_button = SmartButton(self, name='editbtn', text='Edit', command=self.edit, enabled=False)
        self.copy_button = SmartButton(self, name='copybtn', text='Copy', command=self.copy, enabled=False)
        self.delete_button = SmartButton(self, name='deletebtn', text='Delete', command=self.delete, enabled=False)
        
        # Layout the list and the buttons
        self.container.grid(row=0, column=0, columnspan=4, sticky='WE')
        self.new_button.grid(row=1, column=0, sticky='WE')
        self.edit_button.grid(row=1, column=1, sticky='WE')
        self.copy_button.grid(row=1, column=2, sticky='WE')
        self.delete_button.grid(row=1, column=3, sticky='WE')
        
    def highlight(self, widget):
        super(FunctionList, self).highlight(widget)
        self.set_buttons()

    def unhighlight(self):
        super(FunctionList, self).unhighlight()
        self.set_buttons()

    def get_checked(self):
        """Returns the functions that are checked."""
        checked = []
        for child in self.contents.children.values():
            try:
                if child.is_checked():
                    checked.append(child)
            
            # Some children may not support the ``is_checked`` method, so we
            # gracefully ignore them.
            except:
                pass
        
        return checked

    def get_unchecked(self):
        """Returns the functions that are unchecked."""
        unchecked = []
        for child in self.contents.children.values():
            try:
                if not child.is_checked():
                    unchecked.append(child)
            
            # Some children may not support the ``is_checked`` method, so we
            # gracefully ignore them.
            except:
                pass
        
        return unchecked
    
    def set_buttons(self):
        """Sets the button states based on the current selection."""
        self.edit_button.disable()
        self.delete_button.disable()
        self.copy_button.disable()

        function = self.get_selected()
        if function:
            self.delete_button.enable()
            
            try:
                getattr(function, 'window').state()
            except:
                if function.is_editable():
                    self.edit_button.enable()
                    self.copy_button.enable()
                
    
    def clean(self, rows):
        """Allows plain old compiler objects to be passed to the append,
        prepend, set, etc. methods by automagically creating the
        Function objects for them. Yes, this *is* super awesome.
        
        Also forces the rows to be sorted by name. I know that this
        makes append and prepend do exactly the same thing, but tough.
        
        """
        rows = super(FunctionList, self).clean(rows)
        
        cleaned_rows = []
        for row in rows:
            if isinstance(row, (FunctionCompiler, ExternalCompiler)):
                row = Function(self.contents, compiler=row)
                self.bind_row(row)
            cleaned_rows.append(row)
        
        # Sort the compilers by name.
        cleaned_rows.sort(key=lambda x: x.compiler.name)
        
        return cleaned_rows
    
    def update(self, function):
        # Update the functions appearance.
        function.update()
        
        # Lay out all of the functions again, binding all rows, and setting
        # initial backgrounds.
        self.layout()
        self.set_buttons()
        
    def doubleclick_row_callback(self, event):
        """Callback for doubleclicking on an empty row."""
        self.edit()

    def doubleclick_empty_callback(self, event):
        """Callback for doubleclicking on an empty row."""
        self.new()

    def new(self):
        newcompiler = FunctionCompiler(directory=self.directory)
        newcompiler.body = "{\n\n}"
        function = Function(self.contents, compiler=newcompiler)
        self.bind_row(function)
        functionWindow(function, callback=self.update)
        
    def edit(self):
        # TODO: Hacky! Use set_buttons or something.
        # TODO: Closing window should update buttons.
        self.edit_button.disable()

        function = self.get_selected()
        if not function.compiler.is_editable():
            return
        
        try:
            getattr(function, 'window').state()
        except:
            functionWindow(function, callback=self.update) 
        
    def copy(self):
        """Duplicate the selected function."""
        names = [child.compiler.name for child
                 in self.contents.children.values()]
        
        widget = self.get_selected()
        name = widget.compiler.name
        
        i = 1
        newname = name + str(i)
        while newname in names:
            i += 1
            newname = name + str(i)
        
        # Since we can't just deepcopy the whole thing, we have to extract
        # everything and THEN deepcopy it. Fun...
        kwargs = {
            'data': widget.compiler.data,
            'includes': widget.compiler.includes,
            'directory': widget.compiler.directory,
        }
        kwargs = deepcopy(kwargs)
        kwargs['data']['name'] = newname
        
        newcompiler = FunctionCompiler(**kwargs)
        newcompiler.write_source_files()
        
        function = Function(self.contents, compiler=newcompiler)
        self.bind_row(function)
        self.layout()
                
    def delete(self):
        widget = self.get_selected()
        
        # Delete the source files.
        widget.delete()
        
        # Remove the widget from the actual interface.
        self.remove(widget)
        self.set_buttons()

