import time
from tkinter import  Frame, Canvas, Scrollbar, VERTICAL

class HighlightableList(Frame):
    """A list of widgets which implement ``highlight``, ``unhighlight``
    and ``is_highlighted`` methods.
    
    Note that the actually list is encapsulated within a frame within
    the HighlightableList. This allows subclasses to put widgets within
    the HighlightableList so the subclass's methods have access to them.
    
    """
    def __init__(self, *args, **kwargs):
        width = kwargs.get('width', None)
        height = kwargs.get('height', None)
        super(HighlightableList, self).__init__(*args, **kwargs)
        
        self.container = Frame(self)
        self.container.grid()
        
        # Create the Frame's components.
        self.canvas = Canvas(self.container, width=width, height=height, bg='White')
        self.canvas.grid(row=0, column=0, columnspan=2, sticky='NSWE')
        
        # Layout the scrollbar and link it to the canvas.
        self.vscroll = Scrollbar(self.container, orient=VERTICAL,
                                 command=self.canvas.yview)
        self.vscroll.grid(row=0, column=3, sticky='wns')
        self.canvas.config(yscrollcommand=self.vscroll.set)
    
        # Create a subframe and put it in the canvas's scrollable zone.
        self.contents = Frame(self.canvas)
        self.canvas.create_window(0, 0, window=self.contents, anchor='nw')
        
        # TODO: This doesn't work after using changeDir.
        # Bind button clicks to canvas.
        self.canvas.bind('<ButtonRelease-1>', self._empty_callback)
        
        # Configure size of canvas's scrollable zone.
        self.contents.update_idletasks()
        self.canvas.config(scrollregion=(0, 0, width,
                                         self.contents.winfo_height()))
            
    def unhighlight(self):
        """Unhighlight all rows (although only one should ever be
        highlighted at any given time)."""
        for widget in self.contents.children.values():
            try:
                widget.unhighlight()
            except AttributeError:
                pass
    
    def highlight(self, widget):
        """Highlight a specific row."""
        self.unhighlight()
        try:
            widget.highlight()
        except AttributeError:
            pass
    
    def bind_row(self, row):
        row.bind('<ButtonRelease-1>', self._row_callback)
    
    def _row_callback(self, event):
        """Master click callback which sends events out to the click
        and doubleclick callbacks depending on how recently the mouse
        was last clicked. Used for the ***_row_callback functions.
        
        """
        # Get the time the widget was last clicked, and then reset the value.
        last_clicked = getattr(event.widget, 'last_clicked', 0)
        event.widget.last_clicked = time.time()
        
        # Check if whether we should call the click or doubleclick callbacks.
        if time.time() - last_clicked < .5:
            self.doubleclick_row_callback(event)
        else:
            self.click_row_callback(event)
    
    def _empty_callback(self, event):
        """Master click callback which sends events out to the click
        and doubleclick callbacks depending on how recently the mouse
        was last clicked. Used for the ***_empty_callback functions.
        
        """
        # Get the time the widget was last clicked, and then reset the value.
        last_clicked = getattr(event.widget, 'last_clicked', 0)
        event.widget.last_clicked = time.time()
        
        # Check if whether we should call the click or doubleclick callbacks.
        if time.time() - last_clicked < .5:
            self.doubleclick_empty_callback(event)
        else:
            self.click_empty_callback(event)

    def click_row_callback(self, event):
        """Callback for clicking on a row."""
        self.highlight(event.widget)

    def doubleclick_row_callback(self, event):
        """Callback for doubleclicking on an empty row."""
        pass

    def click_empty_callback(self, event):
        """Callback for clicking on an empty row."""
        self.unhighlight()

    def doubleclick_empty_callback(self, event):
        """Callback for doubleclicking on an empty row."""
        pass
    
    # Access and set the internal function objects.
    def clear(self):
        self.set([])
    
    # TODO: Perhaps an add function, a remove function, and a clear function are
    # all we need. All of which can call layout afterwards. Hmm...
    
    def set(self, rows):
        # Get the selected rows index, so we can keep the selection in the same
        # place if this item gets removed.
        selected_index = None
        selected = self.get_selected()
        if selected:
            selected_index = int(selected.grid_info()['row'])
        
        if not isinstance(rows, (list, tuple, set, frozenset)):
            rows = [rows]

        # Clear out any old rows.
        orphans = [child for child in self.contents.children.values()
                   if child not in rows]
        for orphan in orphans:
            orphan.destroy()
        
        # TODO: Clean is being called twice. Here and in layout. How can we make
        # this whole operation smarter.
        
        # Clean the row objects and lay them out in the grid.
        rows = self.clean(rows)
        rows = self.layout()
        
        # Save the selected row, if it still exists, otherwise try to get the
        # closest row to it based on the index of the previously selected row.
        selected = self.get_selected()
        if rows and not selected and selected_index is not None:
            try:
                selected = rows[selected_index]
            except IndexError:
                # We hit the end of the list, so we use the last item.
                selected = rows[-1]

        # Reselect the previously selected row (or as close as we can get).
        if selected:
            self.highlight(selected)
        # We call unhighlight in case a subclass is expecting it to be called
        # when a row is unhighlighted. Note that highlight (above) calls it to.
        else:
            self.unhighlight()
    
    def remove(self, widget):
        rows = [child for child in self.contents.children.values()
                if child is not widget]
        
        self.set(rows)        
    
    def layout(self):
        """Lay out the rows in the given order. If no rows are supplied,
        lays out the existing rows in their current order.
        
        This should be called whenever the items in the list change.
        """
        
        # Get the list of rows.
        rows = list(self.contents.children.values())
        
        # Clean the rows and remove any that shouldn't be layed out.
        clean_rows = []
        for row in self.clean(rows):
            try:
                if row.should_layout():
                    clean_rows.append(row)
                else:
                    row.grid_remove()
            # If the row doesn't support a ``should_layout`` method, we assume
            # the developer wants it to always be included.
            except AttributeError:
                clean_rows.append(row)
        
        for i, row in enumerate(clean_rows):
            row.grid(row=i, sticky='w')
            
        # Update the scrollbar.
        self.contents.update_idletasks()
        self.canvas.config(scrollregion=(0, 0, self.contents.winfo_width(),
                                         self.contents.winfo_height()))
        
        # Bind all the rows again, incase there are new ones.
        for row in clean_rows:
            self.bind(row)
        
        return clean_rows
    
    def clean(self, rows):
        """Cleans all the rows, even existing ones. You can override
        this in a subclass to do cool things, such as sorting by a
        specific row attribute (like a name maybe).
        
        """
        return rows
    
#    def sort(self, rows):
#        """Sorts rows by their grid_info, and ignores any ungrided rows."""
#        return sorted(rows)
        
    def get(self):
        sortable = []
        for row in self.contents.children.values():
            try:
                i = int(row.grid_info()['row'])
                sortable.append(i, row)
            except:
                continue
                    
        # Extract the row objects from the sort.
        rows = zip(*sorted(sortable, cmp=compare))[-1]
        return rows
    
    def get_selected(self):
        """Loops through the children widgets and returns the first
        highlighted one."""

        for child in self.contents.children.values():
            try:
                if child.is_highlighted():
                    return child
            except AttributeError:
                pass
    


class HighlightableRow(Frame):
    """A row in a HighlightableList. Call unhighlight after the object
    has been grided to set the background color for the first time."""
    
    even_row_color = 'White'
    odd_row_color = '#EEEEEE'
    highlight_color = 'LightBlue'
    
    def __init__(self, *args, **kwargs):
        # Make sure the highlight_color is unique, since this is how we detect
        # if a row is highlighted.
        assert self.highlight_color not in (self.even_row_color,
                                            self.odd_row_color,
                                            self.background_color)
        
        super(HighlightableRow, self).__init__(*args, **kwargs)
    
    def _get_widgets_to_highlight(self):
        """Widgets that should have highlighting applied when the row is
        selected."""
        if hasattr(self, '_widgets_to_highlight'):
            return self._widgets_to_highlight
        children = self.children.values()
        return [self] + list(children) 
    def _set_widgets_to_highlight(self, value):
        self._widgets_to_highlight = value
    widgets_to_highlight = property(_get_widgets_to_highlight,
                                    _set_widgets_to_highlight)
    
    def _get_widgets_to_bind(self):
        """Widgets that can be clicked to select the entire row."""
        if hasattr(self, '_widgets_to_bind'):
            return self._widgets_to_bind
        children = self.children.values()
        return [self] + list(children)
    def _set_widgets_to_bind(self, value):
        self._widgets_to_bind = value
    widgets_to_bind = property(_get_widgets_to_bind, _set_widgets_to_bind)
    
    def grid(self, *args, **kwargs):
        super(HighlightableRow, self).grid(*args, **kwargs)
        
        # If the row's not highlighted, we set the default background color,
        # since if it's being re-grided, it's row number has probably changed.
        if not self.is_highlighted():
            self.unhighlight()
    
    @property
    def background_color(self):
        """The default background color for the row."""
        try:        
            # Alternate between the odd and even row colors.
            # NOTE: We add 1 to the row number, since the first row is row 0.
            if (int(self.grid_info()['row']) + 1) % 2:
                return self.odd_row_color
            else:
                return self.even_row_color
        
        # If the row hasn't been grided yet, we just return a default color.
        except:
            return 'White'

    def should_layout(self):
        return True
    
    def highlight(self):
        """Highlights the current row."""
        for widget in self.widgets_to_highlight:
            widget.config(background=self.highlight_color)
    
    def unhighlight(self):
        """Unhighlights the current row."""
        for widget in self.widgets_to_highlight:
            widget.config(background=self.background_color)
        
    def is_highlighted(self):
        """Returns whether this row is highlighted."""
        try:
            # Test whether the first highlight widget is highlighted.
            return (self.widgets_to_highlight[0]['background'] ==
                    self.highlight_color)
        except (IndexError, KeyError):
            return False
        
    def bind(self, event_name, callable, *args, **kwargs):
        """Override bind so that we can pass bind calls onto the
        widget's children. This is how we make the entire row clickable.
        
        """
        # We can't call bind directly on self, since we overrode it.
        if self in self.widgets_to_bind:
            super(HighlightableRow, self).bind(event_name, callable,
                                               *args, **kwargs)
        
        # A wrapper around the callback function, so that event.widget points
        # to HighlightableRow, and not a child widget.
        def wrapper(event):
            while not isinstance(event.widget, HighlightableRow):
                event.widget = event.widget.master
            return callable(event)
            
        for widget in self.widgets_to_bind:
            if widget != self:
                widget.bind(event_name, wrapper, *args, **kwargs)
    
    def unbind(self, event_name, *args, **kwargs):
        """Undo the binding from above."""
        # Since we can't call unbind directly on self.
        if self in self.widgets_to_bind:
            super(HighlightableRow, self).unbind(event_name, *args, **kwargs)
            
        for widget in self.widgets_to_bind:
            if widget != self:
                widget.unbind(event_name, *args, **kwargs)

